/* data.cpp

   Author: marco corvi <marco_corvi@geocities.com>
   Date  : feb 2001

*/

#include "data.h"
#include <iostream>
#include <stdio.h>

int linkNumber = 0;
int pointNumber = 0;
int chainNumber = 0;

/* ------------------------------------------------------------------ */ 
Data::Data( int dx, int dy ) 
  : DIMX(dx), DIMY(dy)
{
  hline = new (unsigned char)[ DIMX*(DIMY+1) ];
  vline = new (unsigned char)[ DIMY*(DIMX+1) ];

  rootChain = NULL;
  rootPoint = NULL;
  rootLink  = NULL;

  reset();
}

Data::~Data() 
{
  delete[] hline;
  delete[] vline;
}

void Data::reset() 
{
  gridlock = 0;
 
  for (int i=0; i< DIMX*(DIMY+1); ++i) 
    hline[i] = EMPTY;
  for (int i=0; i< DIMY*(DIMX+1); ++i) 
    vline[i] = EMPTY;

  freeChainsLinksPoints();

  rootPoint = new (struct Point)[ DIMX*DIMY ];
  for (int i=0; i<DIMX; ++i) for (int j=0; j<DIMY; ++j) {
    int ij = i+j*DIMX;
    rootPoint[ij].order = 0;
    rootPoint[ij].x = i;
    rootPoint[ij].y = j;
    rootPoint[ij].value = EMPTY;
    rootPoint[ij].player = 0;
    rootPoint[ij].unlinked = 0;
    for (int j=0; j<4; ++j)
      rootPoint[i].link[j] = NULL;
    ++ pointNumber ;
  }

  rootLink = new (struct Link)[ 2 * DIMX * DIMY + DIMX + DIMY ];
  for (int j=0; j<=DIMY; ++j) for (int i=0; i<DIMX; ++i) {
    int ij = i + j*DIMX;
    rootLink[ij].x = i;
    rootLink[ij].y = j;
    rootLink[ij].d = HORIZONTAL;
    rootLink[ij].index = linkNumber;
    rootLink[ij].unlinked = 0;
    rootLink[ij].first = rootLink[ij].second = NULL;
    if (j>0) {
      rootLink[ij].first = point(i,j-1);
      add( &rootLink[ij], point(i,j-1) );
    }
    if (j<DIMY) {
      rootLink[ij].second = point(i,j);
      add( &rootLink[ij], point(i,j) );
    }
    rootLink[ij].next = rootLink[ij].prev = NULL;

    ++ linkNumber ;
  }
  for (int i=0; i<=DIMX; ++i) for (int j=0; j<DIMY; ++j) {
    int ij = DIMX*(DIMY+1) + j + i*DIMY;
    rootLink[ij].x = i;
    rootLink[ij].y = j;
    rootLink[ij].d = VERTICAL;
    rootLink[ij].index = linkNumber;
    rootLink[ij].unlinked = 0;
    rootLink[ij].first = rootLink[ij].second = NULL;
    if (i>0) {
      rootLink[ij].first = point(i-1,j);
      add( &rootLink[ij], point(i-1,j) );
    }
    if (i<DIMX) {
      rootLink[ij].second = point(i,j);
      add( &rootLink[ij], point(i,j) );
    }
    rootLink[ij].next = rootLink[ij].prev = NULL;

    ++ linkNumber ;
  }

  for (int i=0; i<DIMX; ++i) {
    for( int j=0; j<= DIMY; ++j) {
      struct Chain * ch1 = insertChain();
      ch1->length = 1;
      ch1->start = ch1->end = link( i, j, HORIZONTAL);
      ch1->start->chain = ch1;
    }
  }

  for(int j=0; j<DIMY; ++j) {
    for (int i=0; i<=DIMX; ++i) {
      struct Chain * ch1 = insertChain();
      ch1->length = 1;
      ch1->start = ch1->end = link( i, j, VERTICAL);
      ch1->start->chain = ch1;
    }
  }

  // dump();
}

int Data::getSquare(int i, int j)
{
  return point(i,j)->value;
}

struct Data::Chain * Data::insertChain()
{
  struct Chain * ch1 = new (struct Chain);
  if (rootChain != NULL) {
    rootChain->prev = ch1;
  }
  ch1->next = rootChain;
  rootChain = ch1;
  ch1->prev = NULL;
  ch1->loop = 0;

  ++ chainNumber ;
  return ch1;
}

void Data::add( struct Link * lk, struct Point * pt )
{
  pt->link[ pt->order ] = lk;
  ++ pt->order;
}

void Data::drop( struct Link * lk, struct Point * pt)
{
  if (pt != NULL) {
    if (lk != NULL) {
      for (int i=0; i<pt->order; ++i) 
        if (pt->link[i] == lk) { pt->link[i]=NULL; break; }
      -- pt->order;
    }
    if (pt->order == 0) unlink( pt );
  }
}





void Data::freeChainsLinksPoints() 
{
  struct Chain * ch;
  struct Chain * ch1 = NULL;
  for (ch=rootChain; ch != NULL; ch = ch1) {
    ch1 = ch->next;
    delete ch;
  }
  rootChain = NULL;

  if (rootPoint != NULL) delete[] rootPoint;
  rootPoint = NULL;

  if (rootLink != NULL) delete[] rootLink;
  rootLink = NULL;
}

struct Data::Point * Data::point(int i, int j) {
  if (i<0 || i>= DIMX || j<0 || j>=DIMY) 
    return NULL;
  return &rootPoint[ i + j*DIMX ];
}

struct Data::Link * Data::link( int i, int j, Direction d) {
  if (d == HORIZONTAL) 
    return &rootLink[ i + j * DIMX ];
  else // d == VERTICAL
    return &rootLink[ DIMX*(DIMY+1) + j + i*DIMY];
}


/* ------------------------------------------------------------------ */
/* "player" should be either 1 or 2
 * returns : 0    no square has been completed
 *           > 0  number of squares completed (either 1 or 2)
 *           < 0  error
 */
int Data::setLine( int i, int j, Direction dir, int player ) 
{
  if (dir == HORIZONTAL && hline[i + j*DIMX] != EMPTY) return -1;
  if (dir == VERTICAL   && vline[j + i*DIMY] != EMPTY) return -1;

  setChainsLinksPoints(i, j, dir);
  // dump();
  // dumpPoints();
  /*
  std::cout << "Chains " << chainNumber 
            << " Points " << pointNumber 
            << " Links " << linkNumber << std::endl;
  */

  int completed = 0;

  if (dir == HORIZONTAL) { 
    hline[i + j*DIMX] = FULL;

    if (j<DIMY) {
      if ( (++ (point(i,j)->value) ) == 4) {
        completed ++;
        point(i,j)->player = player;
        game -> addToPlayer( i, j, player );
      }
    }
    if (j>0) {
      if ( (++ (point(i,j-1)->value)) == 4) {
        completed ++;
        point(i,j-1)->player = player;
        game -> addToPlayer( i, j-1, player );
      }
    }
  } else {         // vertical
    vline[j + i*DIMY] = FULL;

    if (i<DIMX) {
      if ( (++ (point(i,j)->value) ) == 4) {
        completed ++;
        point(i,j)->player = player;
        game -> addToPlayer( i, j, player );
      }
    }
    if (i>0) {
      if ( (++ (point(i-1,j)->value) ) == 4) {
        completed ++;
        point(i-1,j)->player = player;
        game -> addToPlayer( i-1, j, player );
      }
    }
  }

  return completed;
}


int Data::unsetLine( int i, int j, int dir, int player ) 
{
  int completed = 0;

  if (dir == 0) { // horizontal
    if  (hline[i + j*DIMX] == EMPTY) return -1;
    hline[i + j*DIMX] = EMPTY;

    if (j<DIMY) {
      if ( point(i,j)->value -- >= 4) {
        completed ++;
        point(i,j)->player = 0;
        game -> subtractToPlayer( i, j, player );
      }
    }
    if (j>0) {
      if ( point(i,j-1)->value -- >= 4) {
        completed ++;
        point(i,j-1)->player = 0;
        game -> subtractToPlayer( i, j-1, player );
      }
    }
  } else {         // vertical
    if  (vline[ j + i*DIMY] == EMPTY) return -1;
    vline[j + i*DIMY] = EMPTY;

    if (i<DIMX) {
      if ( point(i,j)->value -- >= 4) {
        completed ++;
        point(i,j)->player = 0;
        game -> subtractToPlayer( i, j, player );
      }
    }
    if (i>0) {
      if ( point(i-1,j)->value -- >= 4) {
        completed ++;
	point(i-1,j)->player = 0;
        game -> subtractToPlayer( i-1, j, player );
      }
    }
  }
  return completed;
}

void Data::countChains( int * counts ) 
{
  for (int i=0; i<4; i++) 
    counts[i] = 0;

  for (struct Chain * ch = rootChain; ch != NULL; ch=ch->next) {
    if (ch->length > 3) { ++ counts[3]; }
    else                { ++  counts[ ch->length - 1]; }
  }
  /*
  std::cerr << "Chain counts: " << 
    counts[0] << " " << counts[1] << " " << counts[2] << " " <<
    counts[3] << std::endl;
  */
}
    

void Data::countPoints( int * counts ) 
{
  int i;
  for (i=0; i<5; i++) counts[i] = 0;

  for (i=0; i<DIMX; i++) for (int j=0; j<DIMY; j++)
    counts[ point(i,j)->value ] ++;

  if (! gridlock &&  counts[0] + counts[1] == 0 ) {
    gridlock = 1;
    std::cerr << "gridlock " << std::endl;
  }
  /* 
  std::cerr << "Point counts: " << 
    counts[0] << " " << counts[1] << " " << counts[2] << " " <<
    counts[3] << " " << counts[4];
  */
}

int Data::shortestChainLength( int * count ) 
{
  int len = DIMX*DIMY;

  for (struct Chain * ch = rootChain; ch != NULL; ch = ch->next ) {
    if (ch->length < len) { len = ch->length; *count = 1; }
    else if (ch->length == len) { ++ (*count); }
  }
  
  return len;
}

int Data::getChainLength( int link_index ) 
{
  return rootLink[ link_index].chain->length;
}

int Data::getNearLink( int link_index )
{
  struct Link * lk1 = &(rootLink[ link_index ]);
  struct Chain * ch = lk1->chain;
  if (lk1 == ch->end) {
    lk1 = lk1->prev;
  } else {
    lk1 = lk1->next;
  }
  return lk1->index;
 
}
  


void Data::getChainLinks( int len, int c, int * x, int * y, int * d ) 
{
  std::cerr << "Data: get " << c << " chain of length " << len << std::endl;
  struct Chain * ch;
  for (ch = rootChain; ch != NULL; ch = ch->next ) {
    if (ch->length == len) --c ;
    if (c == 0) break;
  }

  if (ch == NULL) {
    std::cerr << "Error Not enough chains of length " << len << std::endl;
    return;
  }

  struct Link * lk;
  for (lk = ch->start; lk != ch->end; lk = lk->next) {
    *x++ = lk->x;
    *y++ = lk->y;
    *d++ = lk->d;
  }
  *x++ = lk->x;
  *y++ = lk->y;
  *d++ = lk->d;
}



/*  ---------------------------------------------------------------- */
void Data::unlink( struct Point * pt ) { 
  if (pt->unlinked == 1) 
     printf("Unlink POINT %d %d \n", pt->x, pt->y );
  pt->unlinked = 1; 
  -- pointNumber; 
}

void Data::unlink( struct Link * lk ) { 
  if (lk->unlinked == 1) 
    printf("Unlink LINK %d %d %c\n", 
      lk->x, lk->y, (lk->d == HORIZONTAL)? 'H' : 'V');
  lk->unlinked = 1;
  -- linkNumber; 
}

void Data::unlink( struct Chain * ch ) 
{
  // printf("Unlink CHAIN %x (length %d)\n", (unsigned int)ch, ch->length);
  if (ch != NULL) {
    if (ch->prev != NULL) { ch->prev->next = ch->next; }
    else                  { rootChain = ch->next; }

    if (ch->next != NULL) ch->next->prev = ch->prev;
    delete ch;
    -- chainNumber ;
  }
}

void Data::reverse( struct Chain * ch ) 
{
  struct Link * lk1;
  struct Link * lk2;

  for (lk1 = ch->start; lk1 != NULL; lk1=lk2) {
    lk2 = lk1->next;
    lk1->next = lk1->prev;
    lk1->prev = lk2;
  }
  lk1 = ch->start;
  ch->start = ch->end;
  ch->end = lk1;
}

void Data::mergeChains( struct Point * pt ) 
{
  struct Chain * ch1;
  struct Chain * ch2;
  struct Link * lk1;
  struct Link * lk2;

  int i;
  for (i=0; i<4; ++i) 
    if ( pt->link[i] != NULL) { lk1=pt->link[i]; break; }
  for (i++; i<4; ++i) 
    if ( pt->link[i] != NULL) { lk2=pt->link[i]; break; }

  ch1 = lk1->chain;
  ch2 = lk2->chain;

  if (ch1 == ch2) { // make chain into a loop
    ch1->start->prev = ch1->end;
    ch1->end->next   = ch1->start;
    ch1->loop = 1;
  } else {
    if (lk1 != ch1->end) reverse( ch1 );  
    if (lk2 != ch2->start) reverse( ch2 );  

    lk1 = ch1->end;
    lk2 = ch2->start;
    lk1->next = lk2;
    lk2->prev = lk1;
    ch1->end = ch2->end;
    for (; lk2 != NULL; lk2=lk2->next) 
      lk2->chain = ch1;

    ch1->length += ch2->length;
    unlink( ch2 );
  }
}

void Data::splitChain( struct Link * lk, struct Chain * ch )
{
  if (ch->loop == 1) {
    ch->start = lk->next;
    ch->end   = lk->prev;
    ch->start->prev = NULL;
    ch->end->next   = NULL;
    -- ch->length;
    ch->loop = 0;
  } else {
    struct Chain * ch2 = insertChain();
    struct Link * lk1 = lk->prev;
    struct Link * lk2 = lk->next;

    ch2->start = lk2;
    ch2->end = ch->end;
    lk2->prev = NULL;

    ch->end = lk1;
    lk1->next = NULL;
  
    int i=0;
    for (; lk2!=NULL; lk2=lk2->next) { lk2->chain = ch2; ++i; }
    ch2->length = i;
    ch->length -= i + 1;
  }

  unlink( lk );
  
}

void Data::drop( struct Link * lk, struct Chain * ch ) 
{
  if (lk == ch->start) {
    if ( (ch->start = lk->next) != NULL) ch->start->prev = NULL;
  } 
  if (lk == ch->end) {
    if ( (ch->end = lk->prev) != NULL) ch->end->next = NULL;
  }
  --  ch->length ;
  if (ch->length <= 0) unlink( ch );
  unlink(lk);
}

void Data::setChainsLinksPoints( int i, int j, Direction d )
{
  int s1, s2;
  int i1 = i;
  int i2 = i;
  int j1 = j;
  int j2 = j;

  if (d == HORIZONTAL) {
    j2 = j-1;
  } else {
    i2 = i-1;
  }

  struct Link * lk = link( i, j, d );
  if (i1<DIMX && j1<DIMY) {
    s1 = 4 - point(i1,j1)->value;
  } else {
    s1 = 0;
  }
  if (i2 >= 0 && j2 >= 0) { 
    s2 = 4 - point(i2,j2)->value;
  } else {
    s2 = 0;
  }
  drop( lk, point(i1, j1) );
  drop( lk, point(i2, j2) );
  if (s1 == 2 && s2 == 2 ) {
    splitChain( lk, lk->chain );
  } else {
    drop( lk, lk->chain );
    if (s1 == 3) mergeChains( point(i1,j1) );  
    if (s2 == 3) mergeChains( point(i2,j2) );  
  }
}

/* ################################################################# */
/* DEBUG only                                                        */
/*                                                                   */   
void Data::dumpChain( struct Chain * ch ) 
{
  struct Link * lk;

  if (ch->loop == 1) {
    printf("=%x: ", (unsigned int)ch & 0xffff);
  } else {
    printf(" %x: ", (unsigned int)ch & 0xffff);
  }
  for (lk = ch->start; lk != ch->end; lk = lk->next) {
    printf("%d %d %c ", lk->x, lk->y, (lk->d == HORIZONTAL) ? 'H' : 'V' );
    if (lk->chain != ch) {
      printf("\nError link %x Chain %x mismatch\n", 
        (unsigned int)lk & 0xffff, (unsigned int)ch & 0xffff);
    }
  }
  printf("%d %d %c\n", lk->x, lk->y, (lk->d == HORIZONTAL) ? 'H' : 'V' );
}

void Data::dump() 
{  
  struct Chain * ch;

  for (ch = rootChain; ch != NULL; ch = ch->next) 
    dumpChain( ch );
}  

void Data::dumpPoints( ) 
{
  int i, j;
  for (j=0; j<DIMY; j++) {
    printf("Row %d:  ", j);
    for (i=0; i<DIMX; i++) {
      struct Point * pt = &rootPoint[i+j*DIMX];
      printf(" %d-%d ", pt->order, pt->value);
    }
    printf("\n");
  }
}
