/* data.h

   Author: marco corvi
   Date  : jan 2001

*/
#ifndef DATA_H
#define DATA_H

#include "game.h"

class Data {
  public:
    enum { EMPTY=0, ONE, TWO, THREE, FULL };
    enum Direction { HORIZONTAL, VERTICAL };

    Data( int dx, int dy);
    ~Data();

    void reset();
      /* "player" should be either 1 or 2
       * returns : 0    no square has been completed
       *           > 0  number of squares completed (either 1 or 2)
       */
    int setLine( int i, int j, Direction dir, int player );
    int unsetLine( int i, int j, int dir, int player );
    void countPoints( int * counts );
    void countChains( int * counts );

    int getSquare( int i, int j );
    int getHLine( int i, int j ) { return hline[i + j*DIMX]; }
    int getVLine( int i, int j ) { return vline[j + i*DIMY]; }

    void setGame( Game * game ) { this->game = game; }

    bool isGridlock() { return gridlock == 1; }

    void getChainLinks( int len, int c, int * x, int * y, int * d );
    int shortestChainLength( int * count );
    int getChainLength( int link_index );
    int getNearLink( int link_index );

  private:
    int DIMX;
    int DIMY;
    int gridlock;

    Game * game;

    unsigned char * hline;
    unsigned char * vline;

    struct Chain;
    struct Link;
    struct Point {
      int x, y;
      int order;
      int value;
      int player;
      int unlinked;
      struct Data::Link * link[4];
    };

    struct Chain {
      int length;
      int loop;
      struct Data::Link  * start;  // start link
      struct Data::Link  * end;    // end link
      struct Data::Chain * next;   // next chain in the list of chains
      struct Data::Chain * prev;   // prev chain 
    };

    struct Link {
      int x, y;
      int index;                   // offset in the link array
      Direction d;
      int unlinked;
      struct Data::Point * first;  // point on one side of the link
      struct Data::Point * second; // point on the other side
      struct Data::Chain * chain;  // chain that the link belongs to
      struct Data::Link * next;    // next link in the chain
      struct Data::Link * prev;    // prev link in the chain
    };
 

    struct Point * rootPoint;
    struct Chain * rootChain;
    struct Link  * rootLink;

    struct Link * link( int i, int j, Direction dir );
    struct Point * point( int i, int j);
    void add( struct Link * lk, struct Point * pt);
    void drop( struct Link * lk, struct Point * pt);
    void drop( struct Link * lk, struct Chain * ch);
    void freeChainsLinksPoints();
    struct Chain * insertChain();

    void reverse( struct Chain * );
    void unlink( struct Chain * );
    void unlink( struct Point * );
    void unlink( struct Link * );

    void splitChain( struct Link *, struct Chain * );
    void mergeChains( struct Point * );

    void setChainsLinksPoints( int i, int j, Direction dir );
 
public:
    void dump();
    void dumpPoints();
    void dumpChain( struct Chain * );

};

#endif // DATA_H
