/* gui.cpp

   Graphical User Interface (GUI) for the game "square"

   Author: marco corvi
   Date  : jan 2001

*/

#include "gui.h"
#include <iostream>
#include <stdio.h>
#include <stdlib.h>  /* calloc */

void Gui::drawLine( int i, int j, int dir, int color) 
{
  int x1,y1,x2,y2;

  x1 = DXZERO + i * DX;
  y1 = DYZERO + j * DY;
  XSetForeground(theDisplay, theGC, theColor[color].pixel);
  if (dir == 0) {
    x2 = x1 + DX-2;
    x1 += 2;
    XDrawLine(theDisplay, theWindow, theGC, x1, y1, x2, y1);
    ++ y1;
    XDrawLine(theDisplay, theWindow, theGC, x1, y1, x2, y1);
  } else {
    y2 = y1 + DY - 2;
    y1 += 2;
    XDrawLine(theDisplay, theWindow, theGC, x1, y1, x1, y2);
    ++ x1;
    XDrawLine(theDisplay, theWindow, theGC, x1, y1, x1, y2);
  }
}

void Gui::drawSquare( int i, int j, int player) 
{
  int color = (player == 0) ? 0 : 3 + player;

  int x1 = DXZERO + i * DX + 6;
  int y1 = DYZERO + j * DY + 6;
  int x2 = DXZERO + (i+1) * DX - 6;
  int y2 = DYZERO + (j+1) * DY - 6;

  XSetForeground(theDisplay, theGC, theColor[color].pixel);
  XDrawLine(theDisplay, theWindow, theGC, x1, y1, x1, y2);
  XDrawLine(theDisplay, theWindow, theGC, x1, y2, x2, y2);
  XDrawLine(theDisplay, theWindow, theGC, x2, y1, x2, y2);
  XDrawLine(theDisplay, theWindow, theGC, x1, y1, x2, y1);
  XDrawLine(theDisplay, theWindow, theGC, x1, y1, x2, y2);
  XDrawLine(theDisplay, theWindow, theGC, x1, y2, x2, y1);
}

void Gui::drawScores(int count, int * sc, int player) 
{
  char score[12];

  sprintf(score, "Move  %d", count);

  XSetForeground(theDisplay, theGC, theColor[ 3 ].pixel);
  XDrawImageString(theDisplay, theWindow, theGC, 
      guiScore[0].x, guiScore[0].y+10, score, strlen(score) );

  for (int i=0; i<2; i++) {
    sprintf(score, "Player-%1d  %d", i+1, sc[i]);
    XSetForeground(theDisplay, theGC, theColor[ 3 + 1 + i ].pixel);
    XDrawImageString(theDisplay, theWindow, theGC, 
      guiScore[i+1].x, guiScore[i+1].y+10, score, strlen(score) );
  }
  XFlush(theDisplay);
}

void Gui::drawWinner(int player, char * msg) 
{
  XSetForeground(theDisplay, theGC, theColor[3+player].pixel);
  Font theFont = XLoadFont(theDisplay, "lucidasans-18");
  XSetFont(theDisplay, theGC, theFont);
  XDrawImageString(theDisplay, theWindow, theGC, 
      (wW-10*strlen(msg))/2, wH-30, msg, strlen(msg) );
  XFlush(theDisplay);
  // theFont    = XLoadFont(theDisplay, "lucidasans-10");
  // XSetFont(theDisplay, theGC, theFont);
}

void Gui::reset() 
{
  XClearWindow(theDisplay, theWindow);         // clear the game window
  drawTitle();
  drawMenu();
  drawPointsAndLines();
}

/* ------------------------------------------------------------------ */
/* mouseClick returns the indices where
   ii1 : the mouse button has been pressed
   ii2 : the mouse button has been released
   It returns only when the mouse has been pressed and released 
*/
int Gui::mouseClick() 
{
  int i1, i2 ;
  XEvent    event1, event2;   // a couple of X-events

  i2 = -1;
  do { 
    XNextEvent(theDisplay, &event1);
    if (event1.type == ButtonPress) {
      XNextEvent(theDisplay, &event2); 
      while (event2.type != 5) XNextEvent(theDisplay, &event2);
      i1 = getIndex(event1.xbutton.x, event1.xbutton.y);
      i2 = getIndex(event2.xbutton.x, event2.xbutton.y);
    } 
  } while (i2 != i1);
  return(i2);
}

/* ------------------------------------------------------------------ */ 
/* PRIVATE METHODS                                                    */
/* ------------------------------------------------------------------ */ 
void Gui::drawPointsAndLines() 
{
  int i,j;

  XSetForeground(theDisplay, theGC, theColor[ 1 ].pixel);
  for (i=0; i<=DIMX; i++) {
    for (j=0; j<=DIMY; j++) {
      drawPoint( i, j);
    }
  }
  for (i=0; i<=DIMX; i++) {
    for (j=0; j<=DIMY; j++) {
      if (i<DIMX) drawLine( i, j, 0, 2);
      if (j<DIMY) drawLine( i, j, 1, 2);
    }
  }
}

void Gui::drawPoint( int i, int j) 
{
  int x,y;

  x = DXZERO + i * DX;
  y = DYZERO + j * DY;
  XDrawPoint(theDisplay, theWindow, theGC, x-1, y-1);
  XDrawPoint(theDisplay, theWindow, theGC, x,   y-1);
  XDrawPoint(theDisplay, theWindow, theGC, x+1, y-1);
  XDrawPoint(theDisplay, theWindow, theGC, x-1, y);
  XDrawPoint(theDisplay, theWindow, theGC, x,   y);
  XDrawPoint(theDisplay, theWindow, theGC, x+1, y);
  XDrawPoint(theDisplay, theWindow, theGC, x-1, y+1);
  XDrawPoint(theDisplay, theWindow, theGC, x,   y+1);
  XDrawPoint(theDisplay, theWindow, theGC, x+1, y+1);
}

void Gui::drawMenu() 
{
  for (int i=0; i<nMenu; i++) {
    XDrawImageString(theDisplay, theWindow, theGC, 
      guiMenu[i].x, guiMenu[i].y+10, guiMenu[i].name, strlen(guiMenu[i].name) );
  }
  XFlush(theDisplay);
}

void Gui:: drawTitle() 
{
  XStoreName(theDisplay, theWindow, "  SQUARE  ");
  XFlush(theDisplay);
}

      
void Gui::initLayout() 
{
  int i;

  for (i=0; i<nMenu; i++) {                    // init the menu
    guiMenu[i].x = GAP+ (MX+GAP)*i;
    guiMenu[i].y = MYZERO;
  }
  strcpy(guiMenu[0].name, " New Game ");
  strcpy(guiMenu[1].name, " Restart  ");
  strcpy(guiMenu[2].name, " Backward ");
  strcpy(guiMenu[3].name, " Forward  ");
  strcpy(guiMenu[4].name, " Quit     ");
  for (i=0; i<3; i++) {                    // init the menu
    guiScore[i].x = GAP+ (MX+GAP)*i;
    guiScore[i].y = 6*MYZERO;
  }
}

/* --------------------------------------------------------------------- */
/* The function getIndex() return the index of the "segment" in the
   window to which the point (x1,y1) belongs:
*/
int Gui::getIndex(int x1, int y1) 
{
  int y2, x2;
  int i;
  
  if (y1 > DYZERO - 15) {
    x2 = (x1 - DXZERO) % DX;
    y2 = (y1 - DYZERO) % DY;
    x1 = (x1 - DXZERO) / DX;
    y1 = (y1 - DYZERO) / DY;
    if (y2 < DY/4 || DY-y2 < DY/4) {
      if (DY-y2 < DY/4) y1++;
      if (x1 < DIMX && x1 >= 0 && y1 <= DIMY && y1 >= 0)
        return ( x1 + y1 * DIMX );
    } else if ( x2 < DX/4 || DX-x2 < DX/4) {
      if (DX-x2 < DX/4) x1++;
      if (x1 <= DIMX && x1 >= 0 && y1 < DIMY && y1 >= 0)
        return ( DIMX*(DIMY+1) + y1 + x1 * DIMY );
    } else {
      return ( 2*DIMX *DIMY + DIMX + DIMY );
    }
  } 

  for (i=0; i<nMenu; i++) {
    if ((x1>guiMenu[i].x) && (x1<guiMenu[i].x + MX) &&
      (y1>guiMenu[i].y) && (y1<guiMenu[i].y + MY) )
      break;
  }
  return ( -1-i );
}

/* ------------------------------------------------------------------ */ 
Gui::Gui(int dx, int dy)
  : DIMX(dx), DIMY(dy)
{
  wW = DIMX * DX + 2 * DXZERO;
  wH = DIMY * DY + 2 * DYZERO;
  if (wW < 5*DX) wW = 5*DX;
  if (wH < 3*DY) wH = 3*DY;

  int numColors = 6;
  char * colorStr[] =
      {"black", "white", "gray", "red", "green", "blue", "yellow" };

  theDisplay = XOpenDisplay(XDisplayName(NULL));
  if (theDisplay == NULL) {
    std::cerr << "You need to run the program under X-window.\n\n";
    _exit(1);
  }
  int theScreen      = XDefaultScreen(theDisplay);
  Window rootW       =  RootWindow(theDisplay, theScreen);
  theGC      = XDefaultGC(theDisplay, theScreen);
  Colormap theCmap   = XDefaultColormap(theDisplay, theScreen);
  Visual * theVisual =  DefaultVisual(theDisplay, theScreen);
  int theDepth       = XDefaultDepth(theDisplay, theScreen);

  // Screen * pScreen   = XDefaultScreenOfDisplay(theDisplay);
  Font theFont       = XLoadFont(theDisplay, "lucidasans-10");
  XSetFont(theDisplay, theGC, theFont);

  for (int i = 0; i < numColors; i++) {
    if ( XAllocNamedColor(theDisplay, theCmap, 
           colorStr[i], &theColor[i], &theColor[i]
       ) == 0) {
      std::cerr << "Unable to allocate requested color " << colorStr[i];
      std::cerr << "Using white.\n";
      XAllocNamedColor(theDisplay, theCmap, 
           "white", &theColor[i], &theColor[i]);
    }
  }

  XSetBackground(theDisplay, theGC, theColor[ 0 ].pixel);
  XSetForeground(theDisplay, theGC, theColor[ 1 ].pixel);

  XSynchronize(theDisplay, 1);

  XSetWindowAttributes  * xswa = 
    (XSetWindowAttributes *) calloc(1, sizeof( XSetWindowAttributes ) );

  
  // xswa->background_pixmap = 0;
  xswa->background_pixel  = theColor[ 0 ].pixel;
  // xswa->border_pixmap = 0;
  // xswa->border_pixel  = 0;
  // xswa->bit_gravity   = 0;
  // xswa->win_gravity   = 0;
  xswa->backing_store  = Always; /* 2 */
  // xswa->backing_planes = 0;
  // xswa->backing_pixel  = 0;
  xswa->save_under        = 1; // TRUE;
  xswa->event_mask        = ExposureMask | ButtonPressMask | ButtonReleaseMask ;
  xswa->override_redirect = 0; // FALSE;
  xswa->colormap          = theCmap;
  // xswa->cursor            = 0;

  unsigned long flag = 
         CWBackingPlanes | CWDontPropagate | CWSaveUnder | CWBackingStore |
         CWOverrideRedirect | CWEventMask  | CWCursor    | CWBackPixel |
         CWBorderPixel | CWColormap ;

  theWindow = XCreateWindow(theDisplay, rootW, 0, 0, wW, wH, 0, theDepth,
         InputOutput, theVisual, flag, xswa);
  XMapWindow(theDisplay, theWindow);

  initLayout();
  reset();
}

