/* gui.h

   Graphical User Interface (GUI) header for the game "square"

   Author: marco corvi
   Date  : jan 2001

*/

/* -----------------------------------------------------------
   The directory where the xpm pixmap files are:
   (can be defined in the Makefile)
*/
#ifndef GUI_H
#define GUI_H

/* ============== X WINDOW STUFF ==================== */
#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>
#include <X11/X.h>
#include <X11/cursorfont.h>
#include <X11/Xatom.h>

/* ------------------------------------------------------------------ */ 
class Gui {
  public:
    Gui( int dx, int dy);
    void reset();

    void drawLine( int i, int j, int dir, int color);
    void drawSquare( int i, int j, int player);


    void drawScores(int count, int * sc, int player);
    void drawWinner(int player, char * msg);

    int mouseClick();

    /*
    enum { BLACK=0, WHITE, GRAY, RED, GREEN, BLUE, YELLOW };
    */

  private: // methods
    void initLayout();
    int getIndex(int x1, int y1);

    void drawPointsAndLines();
    void drawPoint( int i, int j);
    void drawTitle();
    void drawMenu();

  private: // fields
    int DIMX;
    int DIMY;
    int wW, wH;

    /* ======= LAYOUT SPECIFICATIONS =================== */

    static const int DX     = 75;
    static const int DY     = 75;
    static const int DXZERO = 30;     // horizontal offset
    static const int DYZERO = 70;     // vertical offset

    static const int MYZERO =  5;     // vertical offset of the menu items
    static const int MX     = 75;     // menus' width
    static const int MY     = 15;     // menus' height
    static const int GAP    = 20;

    Display * theDisplay;       // the graphic display
    GC        theGC;            // the graphic context
    Window    theWindow;        // the window for the game 
    XColor    theColor[7];

    /* The menu items: */
    static const int nMenu = 5;

    typedef struct cmenu {
      int x, y;             // item position (x horiz., y vertic.)
      char name[10];        // item label
    } CMENU;
    CMENU  guiMenu[nMenu];
    CMENU  guiScore[3];     // entry 0 for move-number
};


#endif // GUI_H
