/* square.cpp
 *

 */
#include <iostream>
#include <unistd.h> /* usleep getopt */
#include <stdlib.h> /* atoi */
#include <stdio.h>  /* sscanf */

#include "gui.h"
#include "game.h"
#include "data.h"
#include "score.h"
#include "history.h"
#include "util.h"
#include "strategy.h"

#define USER 1
#define COMPUTER 2

#define EXIT -5

#define DX_MAX 9
#define DY_MAX 6
#define DX_DEFAULT 7
#define DY_DEFAULT 5

#define SPEED_UNIT 50000 /* usec */
#define SPEED_DEFAULT 4

#define LEVEL_MIN 1
#define LEVEL_MAX 1

extern char *optarg;
extern int optind, opterr, optopt;

char * OPTARGS = "g:hl:s:";

void doHelp() {
  std::cout <<
  " Dots -n- Boxes \n\n" <<
  " Options: -h    this help\n" <<
  "          -l :  level of the game \n" <<
  "          -s :  speed (0 = fastest, default 4) \n" <<
  "          -g :  geometry (WIDTHxHEIGHT eg 5x4, max 9x6)\n" <<
  std::endl;
}


int main( int argc, char ** argv ) 
{
  int dx = DX_DEFAULT;
  int dy = DY_DEFAULT;

  int speed = SPEED_UNIT * SPEED_DEFAULT;
  int level = LEVEL_MIN;

  int sc[2];

  { int c;

    while (1) {
      c = getopt( argc, argv, OPTARGS );
      if (c == -1) break;
      switch (c) {
        case 'l':
          std::cerr << "Option level " << optarg << std::endl;
          level = atoi( optarg );
          if (level < LEVEL_MIN) level = LEVEL_MIN;
          if (level > LEVEL_MAX) level = LEVEL_MAX;
          break;
        case 's':
          std::cerr << "Option speed " << optarg << std::endl;
          speed = SPEED_UNIT * ( 1 + atoi( optarg ));
          if (speed < SPEED_UNIT) speed = SPEED_UNIT * SPEED_DEFAULT;
          break;
        case 'g':
          sscanf( optarg, "%dx%d", &dx, &dy);
          if (dx > DX_MAX) dx = DX_MAX;
          if (dx < 2)      dx = DX_DEFAULT;
          if (dy > DY_MAX) dy = DY_MAX;
          if (dy < 2)      dy = DY_DEFAULT;
          std::cerr << "Option geometry " << optarg << " " << 
            dx << " " << dy << std::endl;
        case 'h': 
        default:
          doHelp();
          break;
      }
    }
  }
  

  Gui      gui( dx, dy );
  Game     game( dx, dy );
  History  history( dx, dy );
  Score    score;
  Data     data( dx, dy );
  Util     util( dx, dy );
  Strategy strategy( dx, dy );

  game.setGui( & gui ); 
  game.setScore( & score );
  data.setGame( & game );
  strategy.setData( & data );
  strategy.setUtil( & util );

  int i, j, dir, k;
  int player = USER;     // user begins
  int retval;
  int count = 0;

  while (1) {
    if (player == USER) {
      if ( (k = gui.mouseClick( )) == EXIT) break;
    } else {
      k = strategy.doMove();
    }
      
    if ( util.indexToLine( k, &i, &j, &dir) >= 0) {
      Data::Direction dd = ( dir == 0) ? Data::HORIZONTAL : Data::VERTICAL ; 

      if ( (retval = data.setLine( i, j, dd, player)) < 0) {
        std::cerr << " Error: " << i << " " << j << " " << dir << std::endl;
        continue;
      }

      count ++;
      if (player == USER) {
        gui.drawLine( i, j, dir, 3 /* Gui::RED */ );
      } else {
        for (int i0=0; i0<4; i0++) {
          gui.drawLine( i, j, dir, 7 /* Gui::YELLOW */ );
          usleep( speed );
          gui.drawLine( i, j, dir, 0 /* Gui::BLACK */ );
          usleep( speed );
        }
        gui.drawLine( i, j, dir, 3 /* Gui::RED */ );
      }
      score.getScores( sc );
      gui.drawScores( count,  sc,  player );

      if ( history.set( k, player ) ) { // game is over
        break;
      }

      if ( retval ==  0) {
        player = (player == USER) ? COMPUTER : USER;
      }
    } else {
      std::cerr << "Error: Mouse click " << k << std::cerr;
    }
  }

  if (sc[0] > sc[1]) {
      gui.drawWinner(1, "Player 1 won");
  } else if ( sc[1] > sc[0]) {
      gui.drawWinner(2, "Player 2 won");
  } else {
      gui.drawWinner(0, "Players 1 and 2 tied");
  }
  gui.mouseClick( );
  _exit(0);
}




