/* strategy.cpp
 *
 
 */
#include "strategy.h"

#include <stdlib.h>
#include <iostream>

Strategy::Strategy( int dx, int dy )
  : DIMX(dx), DIMY(dy)
{}

int Strategy::doMove() {
  int p_counts[5];
  int c_counts[5];

  data->countPoints( p_counts );
  data->countChains( c_counts );

  std::cerr << "Point: " <<
    p_counts[0] << " " << p_counts[1] << " " << p_counts[2] << " " <<
    p_counts[3] << " " << p_counts[4];
  std::cerr << " Chains: " <<
    c_counts[0] << " " << c_counts[1] << " " << c_counts[2] << " " <<
    c_counts[3] << std::endl ;

  if ( data->isGridlock() ) {
    return doGridlockMove( p_counts, c_counts );
  } else {
    return doNoGridlockMove( p_counts, c_counts );
  }
}

int Strategy::getFourthLine( int x, int y )
{
    if (data->getHLine( x, y ) == Data::EMPTY)
      return util->lineToIndex( x, y, Data::HORIZONTAL );
    if (data->getHLine( x, y+1 ) == Data::EMPTY)
      return util->lineToIndex( x, y+1, Data::HORIZONTAL );
    if (data->getVLine( x, y ) == Data::EMPTY)
      return util->lineToIndex( x, y, Data::VERTICAL );
    if (data->getVLine( x+1, y ) == Data::EMPTY)
      return util->lineToIndex( x+1, y, Data::VERTICAL );
    return -1;
}

int Strategy::doNoGridlockMove( int * p_counts, int * c_counts )
{    
  int k;
  int x = rand() % DIMX;
  int y = rand() % DIMY;

  if ( p_counts[3] > 0 ) {        // 3-side square: get it
    for (int i=0; i<DIMX; ++i) {
      for (int j=0; j<DIMY; ++j) {
        if ( data->getSquare( x, y ) == 3 ) goto done;
        y = (y+1) % DIMY;
      }
      x = (x+1) % DIMX;
    }
done:
    int t;
    if ( (t = getFourthLine( x, y )) >= 0) 
      return t; 

    std::cerr << "Error-1 Square " << x << " " << y << " No line \n";
    data->dumpPoints();

    return -1;

  } else {  // std::cerr << "Strategy: no 3-side squares\n";
    // int toGridlock = p_counts[0] + p_counts[1];
    // int c23 = c_counts[1] + c_counts[2];

    // should make
    //              c23 odd
    //              toGridlock odd (although this is an estimate)
    // i have no idea yet how to implement it
    //

    /* 
     * there are no 3-side squares: 
     * there must be 0- 1- or 2-side squares
     * First try without making 3-side squares
     */
    for (int sides=0; sides < 2; ++ sides) {
      if (p_counts[sides] == 0) continue;

      // std::cerr << "Strategy: " << sides << " side w/o 3 " << std::endl;

      for (int i=0; i<DIMX; ++i) {
        for (int j=0; j<DIMY; ++j) {
          if ( data->getSquare( x, y ) == sides ) {
            if ( (k = trySides( x, y, NO_THREE )) >= 0) {
              /*
              std::cerr << "Strategy: select " << x << " " << y << " index "
                << k << std::endl;
              */
              return k;
            }
          }
          y = (y+1) % DIMY;
        }
        x = (x+1) % DIMX;
      }
    }

    for (int sides=0; sides < 3; ++ sides) {
      if (p_counts[sides] == 0) continue;

      // std::cerr << "Strategy: " << sides << " side(s) with 3 " << std::endl;

      for (int i=0; i<DIMX; ++i) {
        for (int j=0; j<DIMY; ++j) {
          if ( data->getSquare( x, y ) == sides ) {
            if ( (k = trySides( x, y, WITH_THREE )) >= 0) return k;
          }
          y = (y+1) % DIMY;
        }
        x = (x+1) % DIMX;
      }
    }

    std::cerr << "Error-2 Square " << x << " " << y << " No line \n";
    data->dumpPoints();
    return -1;

  }
}

int Strategy::get2_3Chain( int len, int c, int force ) 
{
  int x[3], y[3], d[3];
  data->getChainLinks( len, 1 + rand() % c, x, y, d );
  int i = rand() % len;
  if (len == 3 && force) i = 1;
  std::cerr << "Strategy: len " << len << " index " << i << std::endl;
  return util->lineToIndex( x[i], y[i], d[i] );
}

int Strategy::getShortestChain() 
{
  int c;
  int len = data->shortestChainLength( &c );
  std::cerr << "Strategy: Shortest Chain length " << len << " count " << c
    << std::endl;

  int x[DIMX*DIMY], y[DIMX*DIMY], d[DIMX*DIMY];
  data->getChainLinks( len, 1 + rand() % c, x, y, d );
  int i = rand() % len;

  return util->lineToIndex( x[i], y[i], d[i] );
}

int Strategy::doGridlockMove( int * p_counts, int * c_counts )
{ 
  int c23 = c_counts[1] + c_counts[2];
  int x = rand() % DIMX;
  int y = rand() % DIMY;

  if ( p_counts[3] > 0 ) {        // 3-side square: 
    /* 
     * must change this strategy:
     * if 3-square is in a C2 (2-link 1-box chain) get it
     * if c23 even get the square
     * if c23 == 1 do not get it, but close the domino ???
     */
    int len = 0;
    int t0  = -1;
    int t, l;

    for (int i=0; i<DIMX; ++i) {
      for (int j=0; j<DIMY; ++j) {
        if ( data->getSquare( x, y ) == 3 ) {
          t = getFourthLine( x, y );
          if ( (l=data->getChainLength( t )) > len) {
            len = l;
            t0  = t;
          } 
          if (l == 1)  // one-link chain : get it
            return t;
        }
        y = (y+1) % DIMY;
      }
      x = (x+1) % DIMX;
    }

    if (len > 2)    // if the chain is more than 2 links
      return t0;

    if (c23 == 1) { // get the other link of chain(t0)
      if ( c_counts[3] > 0)
        return data->getNearLink( t0 );
      return t0;
    }
    
    if (c23 & 0x1 == 0)
      return t0;

    /* bad ! get it anyways */
    return t0;

  } else {  
    std::cerr << "Strategy: gridlock no 3-side (" << c23 << ") " 
      << c_counts[1] << " "
      << c_counts[2] << " " << c_counts[3] << std::endl;
    if (c23 % 2 == 0 ) {
      if (c_counts[1] > 0) {                  // get a 2-chain
        return get2_3Chain( 2, c_counts[1], 0 ); 
      } else if (c_counts[2] > 0) {           // get a 3-chain
        return get2_3Chain( 3, c_counts[2], 0 ); 
      } else {               // doomed: pick one from the shortest chain !
        /* get shortest loop if any */
        return getShortestChain( );
      }
    } else {
      if (c_counts[1] > 0) {                   // get a 2-chain
        return get2_3Chain( 2, c_counts[1], 0 ); 
      } else {
        return get2_3Chain( 3, c_counts[2], 1 ); 
      }
    }
  }
  return -1;
}

int Strategy::trySides( int x, int y, int flag )
{ 
  int d;

  static int ioff[4] = {0, 0, 0, 1};  /* line offsets */
  static int joff[4] = {0, 1, 0, 0};
  static int xoff[4] = {0, 0, -1, 1}; /* ngbh square offsets */
  static int yoff[4] = {-1, 1, 0, 0};
  static int dir[4]  = { Data::HORIZONTAL, Data::HORIZONTAL,
                         Data::VERTICAL, Data::VERTICAL };

  int dl, xl, yl, xs, ys;

  int k0 = rand() % 4;
  for (int k=0; k<4; k++) {
    dl = dir[k0];
    xl = x + ioff[k0];
    yl = y + joff[k0];
    xs = x + xoff[k0];
    ys = y + yoff[k0];

    if ( dl == Data::HORIZONTAL ) {
      d = data->getHLine( xl, yl );
    } else {
      d = data->getVLine( xl, yl );
    }
    if (d == Data::EMPTY ) {
      if (flag == NO_THREE) {
        if ( xs >= 0 && xs < DIMX && ys >= 0 && ys < DIMY 
             && data->getSquare( xs, ys ) < 2 ) {
          return util->lineToIndex( xl, yl, dl );
        }
      } else {
        return util->lineToIndex( xl, yl, dl );
      }
    }
    k0 = (k0 + 1) % 4;
  }
  return -1;
}

